=begin rdoc

==scriptingutils - some methods useful when writing scripts

Scripts using this module should first call the +init+ method.

=== Copyright

Copyright (C) 2009 Wybo Dekker (<tt>wybo@servalys.nl</tt>)

This program is free software: you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation, either version 3 of
the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but without any warranty; without even the implied warranty of
merchantability or fitness for a particular purpose. See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see http://www.gnu.org/licenses.

=end

require 'optparse'
require 'ostruct'

   MYNAME = File.basename($0)
  WORKDIR = ENV['PWD'] # don't use Dir.getwd - this one is the real workdir
                       # even if called from mk's tempdir
  PATHSEP = File::PATH_SEPARATOR

     NULL = '/dev/null'
   TONULL = ' 2> ' + NULL
 FROMNULL = ' < '  + NULL
 
   DEVTTY = '/dev/tty'
  PRINTER = 'lpr'

class String
  def String.colortype(c=nil)
    @@colortype = c
    case c
    when :ansi
      @@colors = {
        :err => '[31;1m',
        :fil => '[33m',
        :com => '[34;1m',
        :lin => '[32;1m',
        :war => '[35;1m',
      }
      @@nocolor = '[0m'
    when :html
      @@colors = {
        :err => '<b><font color="#aa0000">',
        :fil => '<b><font color="#0000aa">',
        :com => '<b><font color="#00aaaa">',
        :lin => '<b><font color="#00aa00">',
        :war => '<b><font color="#aaaa00">',
      }
      @@nocolor = '</font></b>'
    when nil
      @@colors = {
        :err => '',
        :fil => '',
        :com => '',
        :lin => '',
        :war => '',
      }
      @@nocolor = ''
    else
      quit("Illegal colortype #{c}")
    end
  end

  # colorings:
  def err; @@colors[:err] + self + @@nocolor; end
  def fil; @@colors[:fil] + self + @@nocolor; end
  def com; @@colors[:com] + self + @@nocolor; end
  def lin; @@colors[:lin] + self + @@nocolor; end
  def war; @@colors[:war] + self + @@nocolor; end

  # escape tex specials with a backslash
  # example:
  # 'tool_example'.texescape -> 'tool\_example'
  def texescape 
    self.gsub(/([#&$_{}%])/) { '\\'+$1 }.gsub(/([<>])/,'$\1$')
  end
end

class Array
  # escape tex specials in String-type elements with a backslash
  # example:
  # ['tool_example',12].texescape -> ['tool\_example',12]
  def texescape
    self.map { |x| x.class == String ? x.texescape : x }
  end
end

# return prog name, output, success rate, and exit status of a system command
# :call-seq:
#   h = sys(command)
#   h[:progname] -> name of the external program being run
#   h[:output]   -> output of the program on stdout
#   h[:error]    -> output of the program on stderr
#   h[:succes]   -> true if the program ran successfully, false otherwise
#   h[:status]   -> exit status of the program ($?)
# Use interactive=true if you need interactive IO with the command
# Note that interaction may become needed if errors occur (vpp, mk for example)
def sys(command,interactive=false)
  require 'open3'
  include Open3
  
  if interactive
    system(command)
    status = $?.exitstatus
  else
    stdin,stdout,stderr = popen3(command + '; echo $?')
    out = stdout.readlines.map{ |x| x.chomp }
    err = stderr.readlines.map{ |x| x.chomp }
    prog = command.split.delete_if{ |x| x=~ /=/ }[0]
    status = out.pop.to_i
  end
  success = status == 0
  
  if @verbose
    mess = "system call: ".fil + command.com
    if success
      warn mess + " successful"
    else
      warn mess + " #{'not successful'.err} exit status=#{status.to_s.err}"
      warn <<-EOF.gsub(/^\s+/,'  ')
        progname:   #{prog}
        success:    #{success}
        status:     #{status}
        output:     #{(out||[]).join("\n          ")}
        error:      #{(err||[]).join("\n          ")} 
      EOF
    end
  end
  OpenStruct.new(:progname => prog, :output => out, :error => err, :success => success, :status => status)
end

# executable returns true if an executable _ex_ exists in the PATH;
# if it does not, it returns false, or it dies, depending on _needed_.
def executable(ex,needed=false)
  ex.strip!
  ex =~ /\// and quit("executable (#{ex}) should not contain a /".err,1)
  ENV['PATH'].split(PATHSEP).each do |dir|
    f = File.join(dir,ex)
    return true if File.exist?(f) && File.executable?(f)
  end
  mess = "executable #{ex} needed but not found or not executable".err
  quit(mess,1) if needed
  warn(mess)
  return false
end

# print version and quit
def print_version(release)
  v,d = []
  open($0).each do |x| 
    v,d = $1,$2  if x =~ /\$Id: .* ([\d.]+) (\d+-\d+-\d+)/
  end
  quit("#{MYNAME}: release #{release}, CVS-version #{v} (#{d})")
end

# read an rc file
def read_rc_file(rc,must_exist=false)
  unless rc && File.exist?(rc) && File.readable?(rc)
    quit("rc file #{rc} does not exist or is not readable".err,1) if must_exist
    return
  end
  begin
    open(rc) { |o| eval o.read }
  rescue
    quit("Errors handling rc file #{rc}: #{$!}".err,1)
  end
  @rcfiles.push(rc) # remember which files were read
end

# program initializer: 
# 1. set the defaults (values of _defaults_) for
#    option- and rc-variables (keys of _defaults_)
# 2. read the rc-files
# 3. handle the options, as defined in the _optionproc_ method
# 4. if the --rc option was used, read its arg as an rc-file
# 5. check if all needed executables (in _needed_, maybe nil) are available
def init(needed,optionproc)
  @rcfiles = [] # collect names for a report in case @verbose is true
  @verbose = false
  String.colortype(:ansi)

  # run all standard rc-files
  unless ENV['NORC']
    [ "/etc/#{MYNAME}rc",               # system rc file
      "#{ENV['HOME']}/.#{MYNAME}rc",    # user rc file
      ".#{MYNAME}rc",                   # rc file in working directory
    ].each do |rc|
      read_rc_file(rc)
    end
  end

  # handle options
  help = method(optionproc).call

  # in case a --rc option was used, read its argument as an rc-file:
  read_rc_file(@rc,true) if @rc

  # if the --verbose option was used, report rc-files read:
  if @verbose
    if @rcfiles.size > 0
      warn "These rc files were read:\n   #{@rcfiles.join("\n   ")}"
    else
      warn "No rc files were read"
    end
  end

  # check if all needed executables (in _needed_) are available
  check_execs(needed)
  help
end

# - print a newline, if newline is true, 
# - print the message if not nil, 
# - then exit with the exitvalue
# - print a traceback when exitvalue is between 1 and 10 (reserving higher value for communication between processes)
def quit(mess=nil,exitvalue=0,newline=false)
  warn '' if newline # ensure prompt at start of line
  warn mess if mess
  if exitvalue.between?(1,10)
    for c in caller(0)
      warn c.sub(/#{$0}/,MYNAME) if c =~ /#{MYNAME}/
    end
  end
  # just in case we have chdir-ed into a temporary directory
  # that may have appeared overnight...
  exit(exitvalue)
end

# check the presence of a list of executables
def check_execs(needs)
  needed = case needs
    when Array  then needs
    when String then needs.split
    when Symbol then send(needs)
    when NilClass then []
    else 
      nil
  end
  needed or quit("unsupported class #{needs.class} for needs in check_execs".err,1)
  missing = []
  needed.each do |n|
    n = n.split.shift
    executable(n) or missing.push(n)
  end
  plur = missing.size > 1 ? 's' : ''
  quit("Missing executable#{plur} #{missing.join(" ")}".err,1) if missing.size > 0
end

# extract a TeX log file, showing the most probable error in red and 
# the error's line number in green.
#
# return with:
# - file in which the error was found, plus 
# - the linenumber in that file, and
# - a hash with a key for each file containing an error and/or warnings.
#   the value is an array containing the lines describing the errors/warnings
#   for the file represented by the key.

def texlog_extract(file,warnings_to_skip=nil)
  warnings_to_skip ||= []
  # read all lines in array log, chomp the lines, and add an empty line 
  if File.exist?(file)
    log = open(file).readlines.map { |x| x.chomp }.push("")
  else
    STDERR.puts("No log file (#{file})")
    return
  end           
  
  message = {}
  currentfile = ['']
  message[''] = []
  linenum = 0
  if log[0] !~ /^This is [a-zA-Z]*TeX, Version/
    return nil, 0, { file => "This is probably not a TeX log file".err }
  end
  # undo linebreaks at 80-char lengths
  log2 = []
  while log.size > 0
    log2.push('')
    loop do
      l = log.shift or break
      log2[-1] << l
      break if l.size < 79
    end
  end
  log = log2
  content = ''
  nwarn = p = nerr = 0
  while line = log.shift
    case line
    when /^\s*$/ then next
    when /\(Font\)/ then next
    when /^LaTeX Font Info/ then next
    when /^File:/ then next
    when /warning:\s+/i
      skip = false
      warnings_to_skip.each do |w|
        if line =~ /#{w}/
          skip = true
          break
        end
      end
      unless skip
        nwarn += 1
        message[currentfile[-1]].push(line.sub(/warning:\s+/i,'\&'.war))
        continue = true
        packagename = line =~ /^Package (.*) warning/i ? $1 : 'unknown'
        # some warnings are followed by explanation lines, starting
        # either with whitespace or with the package name in
        # parenthesis+at least 2 spaces:
        while continue
           l = log.shift
           if l =~ /^(\((#{packagename})\))?\s\s+/
             message[currentfile[-1]].push(l.sub(/^(\(.*?\))/) {$1.tr('[(.)A-Za-z]',' ')})
           else 
             continue = false
             log.unshift(l)
           end
        end
      end
    when /^Missing character:/
      nerr += 1
      message[currentfile[-1]].push(line)
    when /^No pages of output/
      nerr += 1
      message[currentfile[-1]].push(line)
    end
  
    # keep track of the file we are in:     
    unless p > 0 # but only up to the first error
      # remove 1: () and 2: things between ()
      line.sub(/\(\)/,'').gsub(/\(\S+?\)/,'').scan(/\(\S+|\)/).each do |f|
        if f == ')'
          currentfile.pop
        else
          f = f.slice(1..-1) # without the initial (
          currentfile.push f
          message[f] ||= []
        end
      end
    end

    # recent texi2dvi uses TeX-option --file-line-error, which
    # replaces the ! with <filename>:<lineno>:
    if line =~ /^!/ or line =~ /^#{currentfile[-1]}:/
      nerr += 1
      if linenum > 0 # this is the second error message
        p = 20
      else # first error's linenumber
        p += 1
        line = line.err
      end
    elsif line =~ /^l\.(\d+) (.*)/ && p > 0 # look for a line number only after an error
      linenum = $1.to_i
      content = $2
      line = line.lin
    end
  
    if p > 0
      message[currentfile[-1]].push(line)
      p += 1
      break if p > 20 # this must be enough to see what's wrong
    end
  end
  if log.index('No pages of output.')
    message[currentfile[-1]].push('No pages of output - is your text body empty?'.war)
    nwarn += 1
  end
  
  if currentfile[-1] and currentfile[-1] != ''
    errorfile = currentfile[-1]
    message[currentfile[-1]].push('file'.err + ': ' + errorfile)
    message[currentfile[-1]].push('line'.err + " #{linenum}: #{content}")
  end

  # check if there is a bibtex log (blg) file and add it if so:
  blg = file.sub(/\.log$/,'.blg')
  if File.exist?(blg)
    m = "From the bibtex log file:\n".err # used only if there are errors
    pr = false # start printing only after a Database line
    bib = nil
    open(blg).readlines.each do |line|
      case line
      when /^The style file:/
        pr = true
      when /Database file.*: (.*)/
        # if we saw already messages for the first .bib file, report only those
        break if pr && bib && !message[bib].empty?
        bib = $1
        currentfile.push(bib)
        message[bib] = []
        next
      when /^You've used/
        break
      else
        if pr
          errorfile = bib
          message[currentfile[-1]].push(m+line.sub(/^Illegal/,'!\&'))
          nerr += 1 unless m.empty?
          m = ''
        end
      end
    end
  end
  message[''] ||= []
  message[''].push("#{nerr} errors #{nwarn} warnings")
  message.delete_if { |k,v| v == [] }
  return errorfile,linenum,(nerr == 0 && nwarn == 0) ? [] : message
end # texlog_extract

# return version info for LaTeX class or style file.
# First argument is the filename, with .cls or .sty extension
# If the second argument is false, a 5-element hash is returned,
#   containing these elements:
#   :filename    => basename of the file
#   :date        => date in yyyy/mm/dd format, relevant for version checking
#   :version     => version of the file
#   :description => description
#   :location    => full path for the file
# If the second argument is true, only the date is returned, as a string in yyyy/mm/dd format
def ltxfileinfo(f,dateonly=false)
  f =~ /(.*)\.(.*)/
  file,ext,sty,cls = $1,$2,'','article'
  case ext
  when 'cls' 
    cls = file
  when 'sty' 
    sty = file
  else
    raise("Need filename with .cls or .sty extension")
  end
  location = `kpsewhich #{f}`.strip
  return nil if location.empty?
  open('|pdflatex --output-directory=/tmp --jobname=ltxfileinfo','w+') do |o|
    o.puts <<-EOF
      \\documentclass{#{cls}}
      \\usepackage{#{sty}}
      \\listfiles
      \\begin{document}
      \\end{document}
    EOF
    o.readlines.each do |l| 
      if l =~ /^ *#{f}/ 
        l = l.strip.split(/\s+/,4)
        if dateonly
          return l[1]
        else
          return {
            :filename    => l[0],
            :date        => l[1],
            :version     => l[2],
            :description => l[3],
            :location    => location
          }
        end
      end
    end
  end
end # ltxfileinfo
# $Id: scriptingutils.rb,v 1.8 2010-06-10 09:40:39 wybo Exp $

